from returns.pipeline import is_successful
from returns.result import Failure, Result, Success

from .architecture import Architecture
from .environment_definitions import DockerImage, ShellExecution
from .experiment import Experiment, ExperimentExecutionInformation, ExperimentStatus
from .nodes import Node
from .pipeline import Pipeline
from .task import Task, TaskDispatcher
from .types import FlagValues

__all__ = [
    "Experiment",
    "ExperimentStatus",
    "ExperimentExecutionInformation",
    "Pipeline",
    "Task",
    "TaskDispatcher",
    "Result",
    "Success",
    "Failure",
    "is_successful",
    "FlagValues",
    "ShellExecution",
    "DockerImage",
    "Node",
    "Architecture",
]



import sys
from typing import Dict, List, Optional

if sys.version_info >= (3, 9):
    from typing import TypeAlias, TypedDict
else:
    from typing_extensions import TypeAlias, TypedDict


ConnectorContext: TypeAlias = Optional[Dict[str, Dict[str, str]]]


class ExecutorsCancellationRequest(TypedDict):
    executors: List[str]
    cancellation_context: ConnectorContext



from __future__ import annotations

import sys
from enum import IntEnum
from typing import Any, Dict, List, Optional, Set, TypedDict, Union

from pydantic import BaseModel
from returns.result import Result

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

NodeProperty = Union[str, float, int, Set[str], None]

TaskElementResult: TypeAlias = Result[Any, Any]
PipelineResult = Dict[str, List[TaskElementResult]]


class NodeRepresentation(TypedDict):
    name: str
    properties: Dict[str, NodeProperty]
    additional_properties: Dict[str, NodeProperty]
    architecture: str


class EnvironmentDefinitionRepresentation(TypedDict):
    commands: List[str]


class DeploymentRepresentation(TypedDict):
    node: NodeRepresentation
    prepared: bool
    executor_id: str
    error: Optional[str]
    pipeline: bytes
    keep_alive_timeout_minutes: int
    cleanup: bool
    environment_definition: EnvironmentDefinitionRepresentation
    environment_definition_type: str


class NodesRepresentation(TypedDict):
    node_pool_type: str
    node_pool_data: List[Union[NodeRepresentation, NodesRepresentation]]


class BuildContextRepresentation(TypedDict):
    python_version: str
    cloudpickle_version: Optional[str]


class RuntimeContextRepresentation(TypedDict):
    ports_mapping: Dict[int, int]
    environment_variables: Dict[str, str]
    additional_arguments: List[str]


class ShellExecutionRepresentation(EnvironmentDefinitionRepresentation):
    runtime_context: RuntimeContextRepresentation


class DockerImageRepresentation(EnvironmentDefinitionRepresentation):
    image: Optional[str]
    build_context: BuildContextRepresentation
    runtime_context: RuntimeContextRepresentation


class ExperimentRepresentation(TypedDict):
    deployment_map: List[DeploymentRepresentation]
    deployment_context: Optional[Dict[str, Dict[str, str]]]


class DeploymentExecutionResultRepresentation(TypedDict):
    node: NodeRepresentation
    pipeline: str
    result: Optional[str]
    error: Optional[str]


class ExperimentExecutionInformationRepresentation(TypedDict):
    status: int
    experiment: Optional[ExperimentRepresentation]
    execution_result: Union[None, str, List[DeploymentExecutionResultRepresentation]]


class PipelineExecutorState(IntEnum):
    LOOKING_FOR_PIPELINE = 0
    EXECUTING = 1
    REPORTING = 2
    FINISHED = 3


class FlagValues(BaseModel):
    text_value: Optional[str] = None
    int_value: int = 0



import logging as __logging
import os as __os
import sys

__logging.basicConfig()

DATABASE_ENDPOINT = __os.environ.get("NETUNICORN_DATABASE_ENDPOINT", None)
DATABASE_ENDPOINT_DEFAULT = False
if DATABASE_ENDPOINT is None:
    DATABASE_ENDPOINT_DEFAULT = True
    DATABASE_ENDPOINT = "127.0.0.1"

DATABASE_USER = __os.environ.get("NETUNICORN_DATABASE_USER", None)
DATABASE_USER_DEFAULT = False
if DATABASE_USER is None:
    DATABASE_USER_DEFAULT = True
    DATABASE_USER = "unicorn"

DATABASE_PASSWORD = __os.environ.get("NETUNICORN_DATABASE_PASSWORD")

DATABASE_DB = __os.environ.get("NETUNICORN_DATABASE_DB", None)
DATABASE_DB_DEFAULT = False
if DATABASE_DB is None:
    DATABASE_DB_DEFAULT = True
    DATABASE_DB = "unicorndb"

__logging.info(f"Connecting to {DATABASE_ENDPOINT}/{DATABASE_DB} as {DATABASE_USER}")

LOGGING_LEVELS = {
    "DEBUG": __logging.DEBUG,
    "INFO": __logging.INFO,
    "WARNING": __logging.WARNING,
    "ERROR": __logging.ERROR,
    "CRITICAL": __logging.CRITICAL,
}

__logger_level_ = __os.environ.get("NETUNICORN_LOG_LEVEL", "INFO").upper()
__logger_level = LOGGING_LEVELS.get(__logger_level_, __logging.INFO)


def get_logger(name: str, level: int = __logger_level) -> __logging.Logger:
    formatter = __logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logger = __logging.getLogger(name)
    logger.handlers.clear()
    logger.propagate = False

    stream_handler = __logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(level)
    logger.addHandler(stream_handler)

    file_handler = __logging.FileHandler(f"{name}.log")
    file_handler.setFormatter(formatter)
    file_handler.setLevel(level)
    logger.addHandler(file_handler)

    logger.setLevel(level)
    logger.info(f"Logger {name} created with level {level}")
    return logger



from __future__ import annotations

import copy
import uuid
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Union

from returns.result import Failure, Success

from .nodes import Node
from .types import PipelineResult

# Keep classes for export
Success = Success
Failure = Failure


class Task(ABC):
    """
    This is a base class for all tasks. All new task classes should inherit from this class.
    The task instance should encapsulate all the logic and data needed to execute the task.
    Task entrypoint is the run() method.
    Task class can have requirements - commands to be executed to change environment to support this task.
    These requirements would be executed with OS shell during environment setup.
    Each task is to be implemented for a specific architecture, platform, or combination (like Linux + arm64).
    TaskDispatcher can be used for selecting a specific task for the given architecture, platform, or combination.

    Task always returns a Result object.
    - If the task's `run` method returns Result object by itself, you'll receive this Result object
    - If the task's `run` method returns any other object, you'll receive a Success with returned_value encapsulated
    - If the task's `run` method fires an exception, you'll receive a Failure with the exception encapsulated
    """

    # Task installation requirements
    requirements: List[str] = []

    # this variable would be overwritten before task start with results of previous tasks
    # format: {task_name: [Result, Result, ...]}
    previous_steps: PipelineResult = {}

    def __init__(self, name: Optional[str] = None) -> None:
        """
        This is a constructor for the task. Any variables (state) that `run` method should use should be provided here.
        Please, do not forget to call `super().__init__()` in your implementation.
        """
        self.name = name or str(uuid.uuid4())  # Each task should have a name
        self.requirements = copy.deepcopy(self.requirements)

    def __call__(self) -> Any:
        return self.run()

    def __repr__(self) -> str:
        type_ = type(self)
        module = type_.__module__
        qualname = type_.__qualname__
        return f"<{module}.{qualname} with name {self.name}>"

    def __str__(self) -> str:
        return self.__repr__()

    def add_requirement(self, command: str) -> Task:
        """
        This method adds a requirement to the local requirements of the task.
        :param command:
        :return:
        """
        self.requirements = copy.deepcopy(
            self.requirements
        )  # make it instance-specific
        self.requirements.append(command)
        return self

    @abstractmethod
    def run(self) -> Any:
        """
        ## This method is to be overridden by your implementation. ##
        This is the entrypoint for the task.
        This method should never have any arguments except `self`. Any arguments that task would use for execution
        should be provided to the constructor and used later by this method.
        This method will always return a Result object. If this method doesn't return a Result object,
        it will be encapsulated into a Result object.
        :return: Result of the execution
        """
        raise NotImplementedError


class TaskDispatcher(ABC):
    """
    This class is a wrapper for several tasks that are designed to implement the same functionality
    but depend on node attributes. Most often you either want to use a specific
    implementation for a specific architecture (e.g., different Tasks for Windows and Linux),
    or instantiate a task with some specific parameters for a specific node (e.g., node-specific IP address).
    You should implement your own TaskDispatcher class and override the dispatch method.

    Dispatching is done by calling the dispatch method that you should implement.
    """

    def __init__(self, name: Optional[str] = None) -> None:
        """
        This is a constructor for the TaskDispatcher. Any variables (state) that `run` method of
        task implementations should use should be provided here.
        Please, do not forget to call `super().__init__()` in your implementation.
        """
        self.name = name or str(uuid.uuid4())  # Each task should have a name

    @abstractmethod
    def dispatch(self, node: Node) -> Task:
        """
        This method takes a node and should return and instance of the task that is designed to be executed on this node.
        The instance could depend on the node information (such as architecture, platform, properties, etc).
        :param node: Node object
        :return: Task object
        """
        raise NotImplementedError


TaskElement = Union[Task, TaskDispatcher]



from __future__ import annotations

import base64
import copy
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Union

from returns.pipeline import is_successful
from returns.result import Result

from .deployment import Deployment
from .nodes import Node, Nodes
from .pipeline import Pipeline
from .types import (
    DeploymentExecutionResultRepresentation,
    ExperimentExecutionInformationRepresentation,
    ExperimentRepresentation,
    PipelineResult,
)
from .utils import LogType


class ExperimentStatus(Enum):
    UNKNOWN = 0
    PREPARING = 1
    READY = 2
    RUNNING = 3
    FINISHED = 4

    def __json__(self) -> int:
        return self.value

    @classmethod
    def from_json(cls, value: int) -> ExperimentStatus:
        return cls(value)


class Experiment:
    def __init__(
        self, deployment_context: Optional[Dict[str, Dict[str, str]]] = None
    ) -> None:
        self.deployment_map: List[Deployment] = []
        self.deployment_context = deployment_context

    def append(self, node: Node, pipeline: Pipeline) -> Experiment:
        self.deployment_map.append(Deployment(node, pipeline))
        return self

    def map(self, pipeline: Pipeline, nodes: Sequence[Node]) -> Experiment:
        if isinstance(nodes, Nodes):
            raise TypeError("Expected sequence of nodes, got Nodes instead")

        for node in nodes:
            if not isinstance(node, Node):
                raise TypeError(f"Expected sequence of nodes, got {type(node)} instead")
            self.append(node, pipeline)
        return self

    def __json__(self) -> ExperimentRepresentation:
        return {
            "deployment_map": [x.__json__() for x in self.deployment_map],
            "deployment_context": self.deployment_context,
        }

    @classmethod
    def from_json(cls, data: ExperimentRepresentation) -> Experiment:
        instance = cls.__new__(cls)
        instance.deployment_map = [
            Deployment.from_json(x) for x in data["deployment_map"]
        ]
        instance.deployment_context = data.get("deployment_context")
        return instance

    def __getitem__(self, item: int) -> Deployment:
        return self.deployment_map[item]

    def __iter__(self) -> Iterator[Deployment]:
        return iter(self.deployment_map)

    def __len__(self) -> int:
        return len(self.deployment_map)

    def __str__(self) -> str:
        return "\n".join([f" - {x}" for x in self.deployment_map])

    def __repr__(self) -> str:
        return self.__str__()

    def __add__(self, other: Experiment) -> Experiment:
        new_map = copy.deepcopy(self)
        new_map.deployment_map.extend(other.deployment_map)
        if other.deployment_context:
            if new_map.deployment_context is None:
                new_map.deployment_context = {}
            new_map.deployment_context.update(other.deployment_context)
        return new_map


class DeploymentExecutionResult:
    def __init__(
        self,
        node: Node,
        serialized_pipeline: bytes,
        result: Optional[bytes],
        error: Optional[str] = None,
    ):
        self.node = node
        self._pipeline = serialized_pipeline
        self._result = result
        self.error = error

    @property
    def pipeline(self) -> Pipeline:
        import cloudpickle

        return cloudpickle.loads(self._pipeline)  # type: ignore

    @property
    def result(
        self,
    ) -> Optional[Tuple[Result[PipelineResult, PipelineResult], LogType]]:
        import cloudpickle

        return cloudpickle.loads(self._result) if self._result else None

    def __str__(self) -> str:
        text = "DeploymentExecutionResult:\n  Node: {self.node}\n"
        if self._result:
            result: Tuple[Result[PipelineResult, PipelineResult], LogType] = self.result  # type: ignore
            text += f"  Result: {type(result[0])}\n"
            if not is_successful(result[0]):
                text += f"   {result[0]}\n"
            else:
                for task_id, task_result in result[0].unwrap().items():
                    text += f"    {task_id}: {task_result}\n"
            text += f"  Logs:\n"
            for line in result[1]:
                text += f"    {line}"
        if self.error:
            text += f"  Error: {self.error}\n"
        text += "\n"
        return text

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> DeploymentExecutionResultRepresentation:
        return {
            "node": self.node.__json__(),
            "pipeline": base64.b64encode(self._pipeline).decode("utf-8"),
            "result": base64.b64encode(self._result).decode("utf-8")
            if self._result
            else None,
            "error": self.error,
        }

    @classmethod
    def from_json(
        cls, data: DeploymentExecutionResultRepresentation
    ) -> DeploymentExecutionResult:
        return cls(
            Node.from_json(data["node"]),
            base64.b64decode(data["pipeline"]),
            base64.b64decode(data["result"]) if data["result"] else None,
            data["error"],
        )


@dataclass(frozen=True)
class ExperimentExecutionInformation:
    status: ExperimentStatus
    experiment: Optional[Experiment]
    execution_result: Union[None, Exception, List[DeploymentExecutionResult]]

    def __str__(self) -> str:
        text = (
            f"ExperimentExecutionInformation:\n"
            f"status: {self.status}\n"
            f"experiment: \n"
            f"{self.experiment}\n"
            f"execution_result:\n"
            f"{self.execution_result}\n"
        )
        return text

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> ExperimentExecutionInformationRepresentation:
        execution_result: Union[
            None, str, List[DeploymentExecutionResultRepresentation]
        ] = None
        if isinstance(self.execution_result, list):
            execution_result = [x.__json__() for x in self.execution_result]
        elif isinstance(self.execution_result, Exception):
            execution_result = str(self.execution_result.__reduce__())
        return {
            "status": self.status.__json__(),
            "experiment": self.experiment.__json__() if self.experiment else None,
            "execution_result": execution_result,
        }

    @classmethod
    def from_json(
        cls, data: ExperimentExecutionInformationRepresentation
    ) -> ExperimentExecutionInformation:
        status = ExperimentStatus.from_json(data["status"])
        experiment = (
            Experiment.from_json(data["experiment"]) if data["experiment"] else None
        )
        execution_result_data = data["execution_result"]
        execution_result: Union[None, Exception, List[DeploymentExecutionResult]] = None
        if execution_result_data:
            if isinstance(execution_result_data, list):
                execution_result = [
                    DeploymentExecutionResult.from_json(x)
                    for x in execution_result_data
                ]
            else:
                execution_result = Exception(execution_result_data)
        return cls(status, experiment, execution_result)
